/*
 * Copyright (C) 2010-2013 Canonical Ltd.
 * Author: Robert Ancell <robert.ancell@canonical.com>
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2 or version 3 of the License.
 * See http://www.gnu.org/copyleft/lgpl.html the full text of the license.
 */

#ifndef LIGHTDM_H_
#define LIGHTDM_H_

#include "lightdm/greeter.h"
#include "lightdm/language.h"
#include "lightdm/layout.h"
#include "lightdm/power.h"
#include "lightdm/session.h"
#include "lightdm/system.h"
#include "lightdm/user.h"

#endif /* LIGHTDM_H_ */
