/*
 * Copyright (C) 2010-2011 Robert Ancell.
 * Author: Robert Ancell <robert.ancell@canonical.com>
 * 
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version. See http://www.gnu.org/copyleft/gpl.html the full text of the
 * license.
 */

#ifndef PRIVILEGES_H_
#define PRIVILEGES_H_

#include <sys/types.h>

void privileges_drop (uid_t uid, gid_t gid);

void privileges_reclaim (void);

#endif /* PRIVILEGES_H_ */
